module libav.libavutil._version;
/*
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

import libav.libavutil.avutil;

/**
 * @file
 * @ingroup lavu
 * Libavutil version macros
 */

/**
 * @defgroup lavu_ver Version and Build diagnostics*
 
 * Macros and function useful to check at compiletime and at runtime
 * which version of libavutil is in use.*
 
 * @{
 */

immutable int LIBAVUTIL_VERSION_MAJOR = 52;
immutable int LIBAVUTIL_VERSION_MINOR = 13;
immutable int LIBAVUTIL_VERSION_MICRO = 0;

immutable int LIBAVUTIL_VERSION_INT   = AV_VERSION_INT(LIBAVUTIL_VERSION_MAJOR, LIBAVUTIL_VERSION_MINOR, LIBAVUTIL_VERSION_MICRO);
immutable string LIBAVUTIL_VERSION    = AV_VERSION(LIBAVUTIL_VERSION_MAJOR, LIBAVUTIL_VERSION_MINOR, LIBAVUTIL_VERSION_MICRO);
alias LIBAVUTIL_VERSION_INT LIBAVUTIL_BUILD;

/**
 * @}*
 
 * @defgroup depr_guards Deprecation guards
 * FF_API_* defines may be placed below to indicate public API that will be
 * dropped at a future version bump. The defines themselves are not part of
 * the public API and may change, break or disappear at any time.*
 
 * @{
 */

immutable bool FF_API_PIX_FMT = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_CONTEXT_SIZE = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_PIX_FMT_DESC = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_AV_REVERSE = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_AUDIOCONVERT = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_CPU_FLAG_MMX2 = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_LLS_PRIVATE = (LIBAVUTIL_VERSION_MAJOR < 53);
immutable bool FF_API_AVFRAME_LAVC = (LIBAVUTIL_VERSION_MAJOR < 53);
/**
 * @}
 */

/* AVUTIL_VERSION_H */
