module libav.libavutil.aes;
/*
 * copyright (c) 2007 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

import libav.libavutil.attributes;
import libav.libavutil._version;

extern (C):

/**
 * @defgroup lavu_aes AES
 * @ingroup lavu_crypto
 * @{
 */

static if(FF_API_CONTEXT_SIZE){
    int av_aes_size;
}

struct AVAES{}

/**
 * Allocate an AVAES context.
 */
AVAES* av_aes_alloc();

/**
 * Initialize an AVAES context.
 * @param key_bits 128, 192 or 256
 * @param decrypt 0 for encryption, 1 for decryption
 */
int av_aes_init(AVAES* a, ubyte* key, int key_bits, int decrypt);

/**
 * Encrypt or decrypt a buffer using a previously initialized context.
 * @param count number of 16 byte blocks
 * @param dst destination array, can be equal to src
 * @param src source array, can be equal to dst
 * @param iv initialization vector for CBC mode, if NULL then ECB will be used
 * @param decrypt 0 for encryption, 1 for decryption
 */
void av_aes_crypt(AVAES* a, ubyte* dst, ubyte* src, int count, ubyte* iv, int decrypt);

/**
 * @}
 */

/* AVUTIL_AES_H */
