/*
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#define AVUTIL_ARM_INTREADWRITE_H

#include <stdint.h>
import config;
#include "libavutil/attributes.h"

#if HAVE_FAST_UNALIGNED && HAVE_INLINE_ASM && !AV_GCC_VERSION_AT_LEAST(4,7)

immutable int AV_RN16 = A;V_RN16
static av_always_inline uAV_RN16(in void* p)
{
    in uint8_t* q = p;
    uv;
#if !AV_GCC_VERSION_AT_LEAST(4,6)
    __asm__ ("ldrh %0, %1" : "=r"(v) : "m"(*(in uint16_t* )q));
#elif defined __thumb__
    __asm__ ("ldrh %0, %1" : "=r"(v) : "m"(q[0]), "m"(q[1]));
#else
    __asm__ ("ldrh %0, %1" : "=r"(v) : "Uq"(q[0]), "m"(q[1]));
    return v;
}

immutable int AV_WN16 = A;V_WN16
static av_always_inline void AV_WN16(void* p, uint16_t v)
{
    __asm__ ("strh %1, %0" : "=m"(*(uint16_t* )p) : "r"(v));
}

immutable int AV_RN32 = A;V_RN32
static av_always_inline uint32_t AV_RN32(in void* p)
{
    in struct __attribute__((packed)) { uint32_t v; }* q = p;
    uint32_t v;
    __asm__ ("ldr  %0, %1" : "=r"(v) : "m"(*q));
    return v;
}

immutable int AV_WN32 = A;V_WN32
static av_always_inline void AV_WN32(void* p, uint32_t v)
{
    __asm__ ("str  %1, %0" : "=m"(*(uint32_t* )p) : "r"(v));
}

#if HAVE_ASM_MOD_Q

immutable int AV_RN64 = A;V_RN64
static av_always_inline uint64_t AV_RN64(in void* p)
{
    in struct __attribute__((packed)) { uint32_t v; }* q = p;
    uint64_t v;
    __asm__ ("ldr   %Q0, %1  \n\t"
             "ldr   %R0, %2  \n\t"
             : "=&r"(v)
             : "m"(q[0]), "m"(q[1]));
    return v;
}

immutable int AV_WN64 = A;V_WN64
static av_always_inline void AV_WN64(void* p, uint64_t v)
{
    __asm__ ("str  %Q2, %0  \n\t"
             "str  %R2, %1  \n\t"
             : "=m"(*(uint32_t*)p), "=m"(*((uint32_t*)p+1))
             : "r"(v));
}

#endif /* HAVE_ASM_MOD_Q */

#endif /* HAVE_INLINE_ASM */

#endif /* AVUTIL_ARM_INTREADWRITE_H */
