module libav.libavutil.acassert;
/*
 * copyright (c) 2010 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * simple assert() macros that are a bit more flexible than ISO C assert().
 * @author Michael Niedermayer <michaelni@gmx.at>
 */

import core.stdc.stdlib;
import libav.libavutil.avutil;
import libav.libavutil.log;

extern (C):

/**
 * assert() equivalent, that is always enabled.
 */
void av_assert0(int cond){
    do {
        if (!cond) {
            av_log(null, AV_LOG_FATAL, cast(char*)"Assertion %s failed at %s:%d\n",
                   AV_STRINGIFY(cond).ptr, __FILE__.ptr, __LINE__);
            abort();
        }
    } while (0);
}

/**
 * assert() equivalent, that does not lie in speed critical code.
 * These asserts() thus can be enabled without fearing speedloss.
 */
debug(1){
     alias av_assert0 av_assert1;
}else{
     void av_assert1(int cond){}
}

/**
 * assert() equivalent, that does lie in speed critical code.
 */
debug(2){
     alias av_assert0 av_assert2;
}else{
     void av_assert2(int cond){}
}

/* AVUTIL_AVASSERT_H */
