module libav.libavutil.bswap;
/*
 * copyright (c) 2006 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * byte swapping routines
 */

import libav.libavutil.attributes;
import libav.libavutil.avconfig;
import std.traits;

static if(HAVE_AV_CONFIG_H){
    static if(ARCH_ARM){
        import libav.libavutil.arm.bswap;
    }else static if(ARCH_AVR32){
       import libav.libavutil.avr32.bswap;
    }else static if(ARCH_BFIN){
       import libav.libavutil.bfin.bswap;
    }else static if(ARCH_SH4){
       import libav.libavutil.sh4.bswap;
    }else static if(ARCH_X86){
       import libav.libavutil.x86.bswap;
    }
}/* HAVE_AV_CONFIG_H */


string AV_BSWAPC(C : char)(C[] s, C[] x){
    return "AV_BSWAP" ~ s ~ "C(" ~ x ~ ")";
}

extern (C):

T AV_BSWAP16C(T)(T x) if(isImplicitlyConvertible!(ushort, T))
{
    return ((x << 8 & 0xff00)  | (x >> 8 & 0x00ff));
}

T AV_BSWAP32C(T)(T x) if(isImplicitlyConvertible!(uint, T))
{
    return (AV_BSWAP16C(x) << 16 | AV_BSWAP16C(x >> 16));
}

T AV_BSWAP64C(T)(T x) if(isImplicitlyConvertible!(ulong, T))
{
    return (AV_BSWAP32C(x) << 32 | AV_BSWAP32C(x >> 32));
}

T av_bswap16(T)(T x) if(isImplicitlyConvertible!(ushort, T))
{
    x= (x>>8) | (x<<8);
    return x;
}

T av_bswap32(T)(T x) if(isImplicitlyConvertible!(uint, T))
{
    return AV_BSWAP32C(x);
}

T av_bswap64(T)(T x) if(isImplicitlyConvertible!(ulong, T))
{
    return cast(ulong)av_bswap32(x) << 32 | av_bswap32(x >> 32);
}

// be2ne ... big-endian to native-endian
// le2ne ... little-endian to native-endian

static if(AV_HAVE_BIGENDIAN){
    ushort av_be2ne16(ushort x){return x;}
    uint av_be2ne32(uint x){return x;}
    ulong av_be2ne64(ulong x){return x;}
    alias av_bswap16 av_le2ne16;
    alias av_bswap32 av_le2ne32;
    alias av_bswap64 av_le2ne64;
    string AV_BE2NEC(C : char)(C[] s, C[] x){return x;}
    alias AV_BSWAPC AV_BE2NEC;
    //
    alias AV_BSWAP16C AV_BE2NE16C;
    alias AV_BSWAP32C AV_BE2NE32C;
    alias AV_BSWAP64C AV_BE2NE64C;
    ushort AV_LE2NE16C(ushort x){return x;}
    uint AV_LE2NE32C(uint x){return x;}
    ulong AV_LE2NE64C(ulong x){return x;}
}else{
    alias av_bswap16 av_be2ne16;
    alias av_bswap32 av_be2ne32;
    alias av_bswap64 av_be2ne64;
    ushort av_le2ne16(ushort x){return x;}
    uint av_le2ne32(uint x){return x;}
    ulong av_le2ne64(ulong x){return x;}
    alias AV_BSWAPC AV_LE2NEC;
    string AV_BE2NEC(C : char)(C[] s, C[] x){return x;}
    //
    ushort AV_BE2NE16C(ushort x){return x;}
    uint AV_BE2NE32C(uint x){return x;}
    ulong AV_BE2NE64C(ulong x){return x;}
    alias AV_BSWAP16C AV_LE2NE16C;
    alias AV_BSWAP32C AV_LE2NE32C;
    alias AV_BSWAP64C AV_LE2NE64C;
}

/* AVUTIL_BSWAP_H */
