module libav.libavutil.error;
/*
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * error code definitions
 */
import core.stdc.errno;
import libav.libavutil.avutil;

extern (C):

/**
 * @addtogroup lavu_error*
 
 * @{
 */


/* error handling */
static if(EDOM > 0){
    int AVERROR(int e){return -e;}   ///< Returns a negative error code from a POSIX error code, to return from library functions.
    int AVUNERROR(int e){return -e;} ///< Returns a POSIX error code from a library function error return value.
}else{
    int AVERROR(int e){return e;}
    int AVUNERROR(int e){return e;}
}

/* Some platforms have E* and errno already negated. */

immutable int AVERROR_DECODER_NOT_FOUND  = (-0x3cbabb08); ///< Decoder not found
immutable int AVERROR_DEMUXER_NOT_FOUND  = (-0x32babb08); ///< Demuxer not found
immutable int AVERROR_ENCODER_NOT_FOUND  = (-0x3cb1ba08); ///< Encoder not found
immutable int AVERROR_EOF                = (-0x5fb9b0bb); ///< End of file
immutable int AVERROR_EXIT               = (-0x2bb6a7bb); ///< Immediate exit was requested; the called function should not be restarted
immutable int AVERROR_FILTER_NOT_FOUND   = (-0x33b6b908); ///< Filter not found
immutable int AVERROR_INVALIDDATA        = (-0x3ebbb1b7); ///< Invalid data found when processing input
immutable int AVERROR_MUXER_NOT_FOUND    = (-0x27aab208); ///< Muxer not found
immutable int AVERROR_OPTION_NOT_FOUND   = (-0x2bafb008); ///< Option not found
immutable int AVERROR_PATCHWELCOME       = (-0x3aa8beb0); ///< Not yet implemented in Libav, patches welcome
immutable int AVERROR_PROTOCOL_NOT_FOUND = (-0x30adaf08); ///< Protocol not found
immutable int AVERROR_STREAM_NOT_FOUND   = (-0x2dabac08); ///< Stream not found
immutable int AVERROR_BUG                = (-0x5fb8aabe); ///< Bug detected, please report the issue
immutable int AVERROR_UNKNOWN            = (-0x31b4b1ab); ///< Unknown error, typically from an external library
immutable int AVERROR_EXPERIMENTAL       = (-0x2bb2afa8); ///< Requested feature is flagged experimental. Set strict_std_compliance if you really want to use it.

/**
 * Put a description of the AVERROR code errnum in errbuf.
 * In case of failure the global variable errno is set to indicate the
 * error. Even in case of failure av_strerror() will print a generic
 * error message indicating the errnum provided to errbuf.*
 
 * @param errnum      error code to describe
 * @param errbuf      buffer to which description is written
 * @param errbuf_size the size in bytes of errbuf
 * @return 0 on success, a negative value if a description for errnum
 * cannot be found
 */
int av_strerror(int errnum, char* errbuf, size_t errbuf_size);

/**
 * @}
 */

/* AVUTIL_ERROR_H */
