module libav.libavutil.intreadwrite;
/+
/*
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

import libav.libavutil.avconfig;
import libav.libavutil.attributes;
import libav.libavutil.bswap;

union av_alias64{
    ulong     u64;
    uint[2]   u32;
    ushort[4] u16;
    ubyte[8]  u8 ;
    double    f64;
    float[2]  f32;
}

union av_alias32{
    uint      u32;
    ushort[2] u16;
    ubyte[4]  u8 ;
    float     f32;
}

union av_alias16{
    ushort    u16;
    ubyte[2]  u8 ;
}

/*
 * Arch-specific headers can provide any combination of
 * AV_[RW][BLN](16|24|32|64) and AV_(COPY|SWAP|ZERO)(64|128) macros.
 * Preprocessor symbols must be defined, even if these are implemented
 * as inline functions.
 */

static if(HAVE_AV_CONFIG_H){
import libav.libavutil.config;
    static if(ARCH_ARM){
        import libav.libavutil.arm/intreadwrite;
    }else static if(ARCH_AVR32){
        import libav.libavutil.avr32/intreadwrite;
    }else static if(ARCH_MIPS){
        import libav.libavutil.mips/intreadwrite;
    }else static if(ARCH_PPC){
        import libav.libavutil.ppc/intreadwrite;
    }else static if(ARCH_TOMI){
        import libav.libavutil.tomi/intreadwrite;
    }else static if(ARCH_X86){
        import libav.libavutil.x86/intreadwrite;
    }
}/* HAVE_AV_CONFIG_H */

/*
 * Map AV_RNXX <-> AV_R[BL]XX for all variants provided by per-arch headers.
 */

static if(AV_HAVE_BIGENDIAN){

    static if (defined(AV_RN16) && !defined(AV_RB16)){
        alias AV_RN16 AV_RB16;
    }else if(!defined(AV_RN16) && defined(AV_RB16)){
        alias AV_RB16 AV_RN16;
    }

    if    defined(AV_WN16) && !defined(AV_WB16)
        define AV_WB16(p, v) AV_WN16(p, v)
    elif !defined(AV_WN16) &&  defined(AV_WB16)
        define AV_WN16(p, v) AV_WB16(p, v)
    }

    if    defined(AV_RN24) && !defined(AV_RB24)
        define AV_RB24(p) AV_RN24(p)
    elif !defined(AV_RN24) &&  defined(AV_RB24)
        define AV_RN24(p) AV_RB24(p)
    }

    if    defined(AV_WN24) && !defined(AV_WB24)
        define AV_WB24(p, v) AV_WN24(p, v)
    elif !defined(AV_WN24) &&  defined(AV_WB24)
        define AV_WN24(p, v) AV_WB24(p, v)
    }

    if    defined(AV_RN32) && !defined(AV_RB32)
        define AV_RB32(p) AV_RN32(p)
    elif !defined(AV_RN32) &&  defined(AV_RB32)
        define AV_RN32(p) AV_RB32(p)
    }

    if    defined(AV_WN32) && !defined(AV_WB32)
        define AV_WB32(p, v) AV_WN32(p, v)
    elif !defined(AV_WN32) &&  defined(AV_WB32)
        define AV_WN32(p, v) AV_WB32(p, v)
    }

    if    defined(AV_RN64) && !defined(AV_RB64)
        define AV_RB64(p) AV_RN64(p)
    elif !defined(AV_RN64) &&  defined(AV_RB64)
        define AV_RN64(p) AV_RB64(p)
    }

    if    defined(AV_WN64) && !defined(AV_WB64)
        define AV_WB64(p, v) AV_WN64(p, v)
    elif !defined(AV_WN64) &&  defined(AV_WB64)
        define AV_WN64(p, v) AV_WB64(p, v)
    }

}else{ /* AV_HAVE_BIGENDIAN */

    if    defined(AV_RN16) && !defined(AV_RL16)
        define AV_RL16(p) AV_RN16(p)
    elif !defined(AV_RN16) &&  defined(AV_RL16)
        define AV_RN16(p) AV_RL16(p)
    }

    if    defined(AV_WN16) && !defined(AV_WL16)
        define AV_WL16(p, v) AV_WN16(p, v)
    elif !defined(AV_WN16) &&  defined(AV_WL16)
        define AV_WN16(p, v) AV_WL16(p, v)
    }

    if    defined(AV_RN24) && !defined(AV_RL24)
        define AV_RL24(p) AV_RN24(p)
    elif !defined(AV_RN24) &&  defined(AV_RL24)
        define AV_RN24(p) AV_RL24(p)
    }

    if    defined(AV_WN24) && !defined(AV_WL24)
        define AV_WL24(p, v) AV_WN24(p, v)
    elif !defined(AV_WN24) &&  defined(AV_WL24)
        define AV_WN24(p, v) AV_WL24(p, v)
    }

    if    defined(AV_RN32) && !defined(AV_RL32)
        define AV_RL32(p) AV_RN32(p)
    elif !defined(AV_RN32) &&  defined(AV_RL32)
        define AV_RN32(p) AV_RL32(p)
    }

    if    defined(AV_WN32) && !defined(AV_WL32)
        define AV_WL32(p, v) AV_WN32(p, v)
    elif !defined(AV_WN32) &&  defined(AV_WL32)
        define AV_WN32(p, v) AV_WL32(p, v)
    }

    if    defined(AV_RN64) && !defined(AV_RL64)
        define AV_RL64(p) AV_RN64(p)
    elif !defined(AV_RN64) &&  defined(AV_RL64)
        define AV_RN64(p) AV_RL64(p)
    }

    if    defined(AV_WN64) && !defined(AV_WL64)
        define AV_WL64(p, v) AV_WN64(p, v)
    elif !defined(AV_WN64) &&  defined(AV_WL64)
        define AV_WN64(p, v) AV_WL64(p, v)
    }

} /* !AV_HAVE_BIGENDIAN */

/*
 * Define AV_[RW]N helper macros to simplify definitions not provided
 * by per-arch headers.
 */

#if defined(__GNUC__) && !defined(__TI_COMPILER_VERSION__)

union unaligned_64 { ulong l; } __attribute__((packed)) av_alias;
union unaligned_32 { uint l; } __attribute__((packed)) av_alias;
union unaligned_16 { ushort l; } __attribute__((packed)) av_alias;

#   define AV_RN(s, p) (((in union unaligned_##s* ) (p))->l)
#   define AV_WN(s, p, v) ((((union unaligned_##s* ) (p))->l) = (v))

#elif defined(__DECC)

#   define AV_RN(s, p) (*((in __unaligned uint##s##_t*)(p)))
#   define AV_WN(s, p, v) (*((__unaligned uint##s##_t*)(p)) = (v))

#elif AV_HAVE_FAST_UNALIGNED

#   define AV_RN(s, p) (((in av_alias##s*)(p))->u##s)
#   define AV_WN(s, p, v) (((av_alias##s*)(p))->u##s = (v))

}else{

#   define AV_RB16(x)                           \
    (((cast(ubyte*)x)[0] << 8) |          \
      (cast(ubyte*)x)[1])
#   define AV_WB16(p, d) do {                   \
        (cast(ubyte*)p)[1] = (d);               \
        (cast(ubyte*)p)[0] = (d)>>8;            \
    } while(0)

#   define AV_RL16(x)                           \
    (((cast(ubyte*)x)[1] << 8) |          \
      (cast(ubyte*)x)[0])
#   define AV_WL16(p, d) do {                   \
        (cast(ubyte*)p)[0] = (d);               \
        (cast(ubyte*)p)[1] = (d)>>8;            \
    } while(0)

#   define AV_RB32(x)                                \
    (((uint)(cast(ubyte*)x)[0] << 24) |    \
               ((cast(ubyte*)x)[1] << 16) |    \
               ((cast(ubyte*)x)[2] <<  8) |    \
                (cast(ubyte*)x)[3])
#   define AV_WB32(p, d) do {                   \
        (cast(ubyte*)p)[3] = (d);               \
        (cast(ubyte*)p)[2] = (d)>>8;            \
        (cast(ubyte*)p)[1] = (d)>>16;           \
        (cast(ubyte*)p)[0] = (d)>>24;           \
    } while(0)

#   define AV_RL32(x)                                \
    (((uint)(cast(ubyte*)x)[3] << 24) |    \
               ((cast(ubyte*)x)[2] << 16) |    \
               ((cast(ubyte*)x)[1] <<  8) |    \
                (cast(ubyte*)x)[0])
#   define AV_WL32(p, d) do {                   \
        (cast(ubyte*)p)[0] = (d);               \
        (cast(ubyte*)p)[1] = (d)>>8;            \
        (cast(ubyte*)p)[2] = (d)>>16;           \
        (cast(ubyte*)p)[3] = (d)>>24;           \
    } while(0)

#   define AV_RB64(x)                                   \
    (((ulong)(cast(ubyte*)x)[0] << 56) |       \
     ((ulong)(cast(ubyte*)x)[1] << 48) |       \
     ((ulong)(cast(ubyte*)x)[2] << 40) |       \
     ((ulong)(cast(ubyte*)x)[3] << 32) |       \
     ((ulong)(cast(ubyte*)x)[4] << 24) |       \
     ((ulong)(cast(ubyte*)x)[5] << 16) |       \
     ((ulong)(cast(ubyte*)x)[6] <<  8) |       \
      (ulong)(cast(ubyte*)x)[7])
#   define AV_WB64(p, d) do {                   \
        (cast(ubyte*)p)[7] = (d);               \
        (cast(ubyte*)p)[6] = (d)>>8;            \
        (cast(ubyte*)p)[5] = (d)>>16;           \
        (cast(ubyte*)p)[4] = (d)>>24;           \
        (cast(ubyte*)p)[3] = (d)>>32;           \
        (cast(ubyte*)p)[2] = (d)>>40;           \
        (cast(ubyte*)p)[1] = (d)>>48;           \
        (cast(ubyte*)p)[0] = (d)>>56;           \
    } while(0)

#   define AV_RL64(x)                                   \
    (((ulong)(cast(ubyte*)x)[7] << 56) |       \
     ((ulong)(cast(ubyte*)x)[6] << 48) |       \
     ((ulong)(cast(ubyte*)x)[5] << 40) |       \
     ((ulong)(cast(ubyte*)x)[4] << 32) |       \
     ((ulong)(cast(ubyte*)x)[3] << 24) |       \
     ((ulong)(cast(ubyte*)x)[2] << 16) |       \
     ((ulong)(cast(ubyte*)x)[1] <<  8) |       \
      (ulong)(cast(ubyte*)x)[0])
#   define AV_WL64(p, d) do {                   \
        (cast(ubyte*)p)[0] = (d);               \
        (cast(ubyte*)p)[1] = (d)>>8;            \
        (cast(ubyte*)p)[2] = (d)>>16;           \
        (cast(ubyte*)p)[3] = (d)>>24;           \
        (cast(ubyte*)p)[4] = (d)>>32;           \
        (cast(ubyte*)p)[5] = (d)>>40;           \
        (cast(ubyte*)p)[6] = (d)>>48;           \
        (cast(ubyte*)p)[7] = (d)>>56;           \
    } while(0)

static if(AV_HAVE_BIGENDIAN){
#   define AV_RN(s, p)    AV_RB##s(p)
#   define AV_WN(s, p, v) AV_WB##s(p, v)
}else{
#   define AV_RN(s, p)    AV_RL##s(p)
#   define AV_WN(s, p, v) AV_WL##s(p, v)

#} /* HAVE_FAST_UNALIGNED */

#   define AV_RN16(p) AV_RN(16, p)

#   define AV_RN32(p) AV_RN(32, p)

#   define AV_RN64(p) AV_RN(64, p)

#   define AV_WN16(p, v) AV_WN(16, p, v)

#   define AV_WN32(p, v) AV_WN(32, p, v)

#   define AV_WN64(p, v) AV_WN(64, p, v)

static if(AV_HAVE_BIGENDIAN){
#   define AV_RB(s, p)    AV_RN##s(p)
#   define AV_WB(s, p, v) AV_WN##s(p, v)
#   define AV_RL(s, p)    av_bswap##s(AV_RN##s(p))
#   define AV_WL(s, p, v) AV_WN##s(p, av_bswap##s(v))
}else{
#   define AV_RB(s, p)    av_bswap##s(AV_RN##s(p))
#   define AV_WB(s, p, v) AV_WN##s(p, av_bswap##s(v))
#   define AV_RL(s, p)    AV_RN##s(p)
#   define AV_WL(s, p, v) AV_WN##s(p, v)

#define AV_RB8(x)     ((cast(ubyte*)x)[0])
#define AV_WB8(p, d)  do { (cast(ubyte*)p)[0] = (d); } while(0)

#define AV_RL8(x)     AV_RB8(x)
#define AV_WL8(p, d)  AV_WB8(p, d)

#   define AV_RB16(p)    AV_RB(16, p)
#   define AV_WB16(p, v) AV_WB(16, p, v)

#   define AV_RL16(p)    AV_RL(16, p)
#   define AV_WL16(p, v) AV_WL(16, p, v)

#   define AV_RB32(p)    AV_RB(32, p)
#   define AV_WB32(p, v) AV_WB(32, p, v)

#   define AV_RL32(p)    AV_RL(32, p)
#   define AV_WL32(p, v) AV_WL(32, p, v)

#   define AV_RB64(p)    AV_RB(64, p)
#   define AV_WB64(p, v) AV_WB(64, p, v)

#   define AV_RL64(p)    AV_RL(64, p)
#   define AV_WL64(p, v) AV_WL(64, p, v)

#   define AV_RB24(x)                           \
    (((cast(ubyte*)x)[0] << 16) |         \
     ((cast(ubyte*)x)[1] <<  8) |         \
      (cast(ubyte*)x)[2])
#   define AV_WB24(p, d) do {                   \
        (cast(ubyte*)p)[2] = (d);               \
        (cast(ubyte*)p)[1] = (d)>>8;            \
        (cast(ubyte*)p)[0] = (d)>>16;           \
    } while(0)

#   define AV_RL24(x)                           \
    (((cast(ubyte*)x)[2] << 16) |         \
     ((cast(ubyte*)x)[1] <<  8) |         \
      (cast(ubyte*)x)[0])
#   define AV_WL24(p, d) do {                   \
        (cast(ubyte*)p)[0] = (d);               \
        (cast(ubyte*)p)[1] = (d)>>8;            \
        (cast(ubyte*)p)[2] = (d)>>16;           \
    } while(0)

/*
 * The AV_[RW]NA macros access naturally aligned data
 * in a type-safe way.
 */

#define AV_RNA(s, p)    (((in av_alias##s*)(p))->u##s)
#define AV_WNA(s, p, v) (((av_alias##s*)(p))->u##s = (v))

#   define AV_RN16A(p) AV_RNA(16, p)

#   define AV_RN32A(p) AV_RNA(32, p)

#   define AV_RN64A(p) AV_RNA(64, p)

#   define AV_WN16A(p, v) AV_WNA(16, p, v)

#   define AV_WN32A(p, v) AV_WNA(32, p, v)

#   define AV_WN64A(p, v) AV_WNA(64, p, v)

/*
 * The AV_COPYxxU macros are suitable for copying data to/from unaligned
 * memory locations.
 */

#define AV_COPYU(n, d, s) AV_WN##n(d, AV_RN##n(s));

#   define AV_COPY16U(d, s) AV_COPYU(16, d, s)

#   define AV_COPY32U(d, s) AV_COPYU(32, d, s)

#   define AV_COPY64U(d, s) AV_COPYU(64, d, s)

#   define AV_COPY128U(d, s)                                    \
    do {                                                        \
        AV_COPY64U(d, s);                                       \
        AV_COPY64U((char* )(d) + 8, (in char* )(s) + 8);     \
    } while(0)

/* Parameters for AV_COPY*, AV_SWAP*, AV_ZERO* must be
 * naturally aligned. They may be implemented using MMX,
 * so emms_c() must be called before using any float code
 * afterwards.
 */

#define AV_COPY(n, d, s) \
    (((av_alias##n*)(d))->u##n = ((in av_alias##n*)(s))->u##n)

#   define AV_COPY16(d, s) AV_COPY(16, d, s)

#   define AV_COPY32(d, s) AV_COPY(32, d, s)

#   define AV_COPY64(d, s) AV_COPY(64, d, s)

#   define AV_COPY128(d, s)                    \
    do {                                       \
        AV_COPY64(d, s);                       \
        AV_COPY64((char*)(d)+8, (char*)(s)+8); \
    } while(0)

#define AV_SWAP(n, a, b) FFSWAP(av_alias##n,* (av_alias##n*)(a),* (av_alias##n*)(b))

#   define AV_SWAP64(a, b) AV_SWAP(64, a, b)


extern(D){
    
    string AV_ZERO(C : char)(C[] n, C[] d)
    {
        return "(av_alias" ~ n ~ "* " ~ d ~ ").u" ~ n ~ " = 0"
    }
    string AV_ZERO(C : char)(C[] n)
    {
        return "(av_alias" ~ n ~ "* " ~ d ~ ").u" ~ n ~ " = 0"
    }
}

alias AV_ZERO("16") AV_ZERO16;
alias AV_ZERO("32") AV_ZERO32;
alias AV_ZERO("64") AV_ZERO64;

void AV_ZERO128(ref av_alias64 d){
{
    do {
        AV_ZERO64(d);
        AV_ZERO64(cast(char*)d+8);
    } while(0)
}

/* AVUTIL_INTREADWRITE_H */
+/