module libav.libavutil.mathematics;
/*
 * copyright (c) 2005 Michael Niedermayer <michaelni@gmx.at>*
 
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

import core.stdc.math;
import libav.libavutil.attributes;
import libav.libavutil.rational;
import libav.libavutil.intfloat;

extern (C):

immutable float M_LOG2_10      = 3.32192809488736234787;  /* log_2 10 */
immutable float M_PHI          = 1.61803398874989484820;   /* phi / golden ratio */
immutable float NAN            = av_int2float(0x7fc00000);
immutable float INFINITY       = av_int2float(0x7f800000);

/**
 * @addtogroup lavu_math
 * @{
 */


enum AVRounding {
    AV_ROUND_ZERO     = 0, ///< Round toward zero.
    AV_ROUND_INF      = 1, ///< Round away from zero.
    AV_ROUND_DOWN     = 2, ///< Round toward -infinity.
    AV_ROUND_UP       = 3, ///< Round toward +infinity.
    AV_ROUND_NEAR_INF = 5, ///< Round to nearest and halfway cases away from zero.
}

/**
 * Return the greatest common divisor of a and b.
 * If both a and b are 0 or either or both are <0 then behavior is
 * undefined.
 */
long av_gcd(long a, long b);

/**
 * Rescale a 64-bit integer with rounding to nearest.
 * A simple a*b/c isn't possible as it can overflow.
 */
long av_rescale(long a, long b, long c);

/**
 * Rescale a 64-bit integer with specified rounding.
 * A simple a*b/c isn't possible as it can overflow.
 */
long av_rescale_rnd(long a, long b, long c, AVRounding);

/**
 * Rescale a 64-bit integer by 2 rational numbers.
 */
long av_rescale_q(long a, AVRational bq, AVRational cq);

/**
 * Rescale a 64-bit integer by 2 rational numbers with specified rounding.
 */
long av_rescale_q_rnd(long a, AVRational bq, AVRational cq, AVRounding);

/**
 * Compare 2 timestamps each in its own timebases.
 * The result of the function is undefined if one of the timestamps
 * is outside the long range when represented in the others timebase.
 * @return -1 if ts_a is before ts_b, 1 if ts_a is after ts_b or 0 if they represent the same position
 */
int av_compare_ts(long ts_a, AVRational tb_a, long ts_b, AVRational tb_b);

/**
 * Compare 2 integers modulo mod.
 * That is we compare integers a and b for which only the least
 * significant log2(mod) bits are known.*
 
 * @param mod must be a power of 2
 * @return a negative value if a is smaller than b
 *         a positive value if a is greater than b
 *         0                if a equals          b
 */
long av_compare_mod(ulong a, ulong b, ulong mod);

/**
 * @}
 */

/* AVUTIL_MATHEMATICS_H */
