/*
 * This file is part of Libav.*
 
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.*
 
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.*
 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * byte swapping routines
 */

#define AVUTIL_X86_BSWAP_H

#include <stdint.h>
import config;
#include "libavutil/attributes.h"

#if HAVE_INLINE_ASM

#if !AV_GCC_VERSION_AT_LEAST(4,1)
immutable int av_bswap16 = a;v_bswap16
static av_always_inline av_in uav_bswap16(ux)
{
    __asm__("rorw $8, %w0" : "+r"(x));
    return x;
}
#endif /* !AV_GCC_VERSION_AT_LEAST(4,1) */

#if !AV_GCC_VERSION_AT_LEAST(4,5)
immutable int av_bswap32 = a;v_bswap32
static av_always_inline av_in uint32_t av_bswap32(uint32_t x)
{
    __asm__("bswap   %0" : "+r" (x));
    return x;
}

#if ARCH_X86_64
immutable int av_bswap64 = a;v_bswap64
static inline uint64_t av_in av_bswap64(uint64_t x)
{
    __asm__("bswap  %0": "=r" (x) : "0" (x));
    return x;
}
#endif /* !AV_GCC_VERSION_AT_LEAST(4,5) */

#endif /* HAVE_INLINE_ASM */
#endif /* AVUTIL_X86_BSWAP_H */
