module libjpeg.jerror;
/*
 * jerror.h
 *
 * Copyright (C) 1994-1997, Thomas G. Lane.
 * Modified 1997-2012 by Guido Vollbeding.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.*
 
 * This file defines the error and message codes for the JPEG library.
 * Edit this file to add new codes, or to translate the message strings to
 * some other language.
 * A set of error-reporting macros are defined too.  Some applications using
 * the JPEG library may wish to include this file to get the error codes
 * and/or the macros.
 */

/*
 * To define the enum list of message codes, include this file without
 * defining macro JMESSAGE.  To create a message string table, include it
 * again with a suitable JMESSAGE definition (see jerror.c for an example).
 */
/* First time through, define the enum list */

/* Repeated inclusions of this file are no-ops unless JMESSAGE is defined */

/* JERROR_H */
/* JMESSAGE */

import libjpeg.jversion;

enum J_MESSAGE_CODE{

  /* JMAKE_ENUM_LIST */

  JMSG_NOMESSAGE, /* Must be first entry! */

  /* For maintenance convenience, list is alphabetical by message code name */
  JERR_BAD_ALIGN_TYPE,
  JERR_BAD_ALLOC_CHUNK,
  JERR_BAD_BUFFER_MODE,
  JERR_BAD_COMPONENT_ID,
  JERR_BAD_CROP_SPEC,
  JERR_BAD_DCT_COEF,
  JERR_BAD_DCTSIZE,
  JERR_BAD_DROP_SAMPLING,
  JERR_BAD_HUFF_TABLE,
  JERR_BAD_IN_COLORSPACE,
  JERR_BAD_J_COLORSPACE,
  JERR_BAD_LENGTH,
  JERR_BAD_LIB_VERSION,
  JERR_BAD_MCU_SIZE,
  JERR_BAD_POOL_ID,
  JERR_BAD_PRECISION,
  JERR_BAD_PROGRESSION,
  JERR_BAD_PROG_SCRIPT,
  JERR_BAD_SAMPLING,
  JERR_BAD_SCAN_SCRIPT,
  JERR_BAD_STATE,
  JERR_BAD_STRUCT_SIZE,
  JERR_BAD_VIRTUAL_ACCESS,
  JERR_BUFFER_SIZE,
  JERR_CANT_SUSPEND,
  JERR_CCIR601_NOTIMPL,
  JERR_COMPONENT_COUNT,
  JERR_CONVERSION_NOTIMPL,
  JERR_DAC_INDEX,
  JERR_DAC_VALUE,
  JERR_DHT_INDEX,
  JERR_DQT_INDEX,
  JERR_EMPTY_IMAGE,
  JERR_EMS_READ,
  JERR_EMS_WRITE,
  JERR_EOI_EXPECTED,
  JERR_FILE_READ,
  JERR_FILE_WRITE,
  JERR_FRACT_SAMPLE_NOTIMPL,
  JERR_HUFF_CLEN_OVERFLOW,
  JERR_HUFF_MISSING_CODE,
  JERR_IMAGE_TOO_BIG,
  JERR_INPUT_EMPTY,
  JERR_INPUT_EOF,
  JERR_MISMATCHED_QUANT_TABLE,
  JERR_MISSING_DATA,
  JERR_MODE_CHANGE,
  JERR_NOTIMPL,
  JERR_NOT_COMPILED,
  JERR_NO_ARITH_TABLE,
  JERR_NO_BACKING_STORE,
  JERR_NO_HUFF_TABLE,
  JERR_NO_IMAGE,
  JERR_NO_QUANT_TABLE,
  JERR_NO_SOI,
  JERR_OUT_OF_MEMORY,
  JERR_QUANT_COMPONENTS,
  JERR_QUANT_FEW_COLORS,
  JERR_QUANT_MANY_COLORS,
  JERR_SOF_BEFORE,
  JERR_SOF_DUPLICATE,
  JERR_SOF_NO_SOS,
  JERR_SOF_UNSUPPORTED,
  JERR_SOI_DUPLICATE,
  JERR_TFILE_CREATE,
  JERR_TFILE_READ,
  JERR_TFILE_SEEK,
  JERR_TFILE_WRITE,
  JERR_TOO_LITTLE_DATA,
  JERR_UNKNOWN_MARKER,
  JERR_VIRTUAL_BUG,
  JERR_WIDTH_OVERFLOW,
  JERR_XMS_READ,
  JERR_XMS_WRITE,
  JMSG_COPYRIGHT,
  JMSG_VERSION,
  JTRC_16BIT_TABLES,
  JTRC_ADOBE,
  JTRC_APP0,
  JTRC_APP14,
  JTRC_DAC,
  JTRC_DHT,
  JTRC_DQT,
  JTRC_DRI,
  JTRC_EMS_CLOSE,
  JTRC_EMS_OPEN,
  JTRC_EOI,
  JTRC_HUFFBITS,
  JTRC_JFIF,
  JTRC_JFIF_BADTHUMBNAILSIZE,
  JTRC_JFIF_EXTENSION,
  JTRC_JFIF_THUMBNAIL,
  JTRC_MISC_MARKER,
  JTRC_PARMLESS_MARKER,
  JTRC_QUANTVALS,
  JTRC_QUANT_3_NCOLORS,
  JTRC_QUANT_NCOLORS,
  JTRC_QUANT_SELECTED,
  JTRC_RECOVERY_ACTION,
  JTRC_RST,
  JTRC_SMOOTH_NOTIMPL,
  JTRC_SOF,
  JTRC_SOF_COMPONENT,
  JTRC_SOI,
  JTRC_SOS,
  JTRC_SOS_COMPONENT,
  JTRC_SOS_PARAMS,
  JTRC_TFILE_CLOSE,
  JTRC_TFILE_OPEN,
  JTRC_THUMB_JPEG,
  JTRC_THUMB_PALETTE,
  JTRC_THUMB_RGB,
  JTRC_UNKNOWN_IDS,
  JTRC_XMS_CLOSE,
  JTRC_XMS_OPEN,
  JWRN_ADOBE_XFORM,
  JWRN_ARITH_BAD_CODE,
  JWRN_BOGUS_PROGRESSION,
  JWRN_EXTRANEOUS_DATA,
  JWRN_HIT_MARKER,
  JWRN_HUFF_BAD_CODE,
  JWRN_JFIF_MAJOR,
  JWRN_JPEG_EOF,
  JWRN_MUST_RESYNC,
  JWRN_NOT_SEQUENTIAL,
  JWRN_TOO_MUCH_DATA,
}

/* JMAKE_ENUM_LIST */

/* Zap JMESSAGE macro so that future re-inclusions do nothing by default */


/* Macros to simplify using the error and trace message stuff */
/* The first parameter is either type of cinfo pointer */

/* Fatal errors (print message and exit) */
/+
#define ERREXIT(cinfo,code)  \
  ((cinfo)->err->msg_code = (code), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
}
#define ERREXIT1(cinfo,code,p1)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT2(cinfo,code,p1,p2)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT3(cinfo,code,p1,p2,p3)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (cinfo)->err->msg_parm.i[2] = (p3), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT4(cinfo,code,p1,p2,p3,p4)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (cinfo)->err->msg_parm.i[2] = (p3), \
   (cinfo)->err->msg_parm.i[3] = (p4), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT6(cinfo,code,p1,p2,p3,p4,p5,p6)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (cinfo)->err->msg_parm.i[2] = (p3), \
   (cinfo)->err->msg_parm.i[3] = (p4), \
   (cinfo)->err->msg_parm.i[4] = (p5), \
   (cinfo)->err->msg_parm.i[5] = (p6), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXITS(cinfo,code,str)  \
  ((cinfo)->err->msg_code = (code), \
   strncpy((cinfo)->err->msg_parm.s, (str), JMSG_STR_PARM_MAX), \
   (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))

#define MAKESTMT(stuff)		do { stuff } while (0)

/* Nonfatal errors (we can keep going, but the data is probably corrupt) */
#define WARNMS(cinfo,code)  \
  ((cinfo)->err->msg_code = (code), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))
#define WARNMS1(cinfo,code,p1)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))
#define WARNMS2(cinfo,code,p1,p2)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))

/* Informational/debugging messages */
#define TRACEMS(cinfo,lvl,code)  \
  ((cinfo)->err->msg_code = (code), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS1(cinfo,lvl,code,p1)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS2(cinfo,lvl,code,p1,p2)  \
  ((cinfo)->err->msg_code = (code), \
   (cinfo)->err->msg_parm.i[0] = (p1), \
   (cinfo)->err->msg_parm.i[1] = (p2), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS3(cinfo,lvl,code,p1,p2,p3)  \
  MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i; \
	   _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); \
	   (cinfo)->err->msg_code = (code); \
	   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS4(cinfo,lvl,code,p1,p2,p3,p4)  \
  MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i; \
	   _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4); \
	   (cinfo)->err->msg_code = (code); \
	   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS5(cinfo,lvl,code,p1,p2,p3,p4,p5)  \
  MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i; \
	   _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4); \
	   _mp[4] = (p5); \
	   (cinfo)->err->msg_code = (code); \
	   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS8(cinfo,lvl,code,p1,p2,p3,p4,p5,p6,p7,p8)  \
  MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i; \
	   _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4); \
	   _mp[4] = (p5); _mp[5] = (p6); _mp[6] = (p7); _mp[7] = (p8); \
	   (cinfo)->err->msg_code = (code); \
	   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMSS(cinfo,lvl,code,str)  \
  ((cinfo)->err->msg_code = (code), \
   strncpy((cinfo)->err->msg_parm.s, (str), JMSG_STR_PARM_MAX), \
   (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
+/
/* JERROR_H */
