module libjpeg.jmorecfg;
/*
 * jmorecfg.h*
 
 * Copyright (C) 1991-1997, Thomas G. Lane.
 * Modified 1997-2012 by Guido Vollbeding.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.*
 
 * This file contains additional configuration options that customize the
 * JPEG software for special applications or support machine-dependent
 * optimizations.  Most users will not need to touch this file.
 */


/*
 * Define BITS_IN_short as either
 *   8   for 8-bit sample values (the usual setting)
 *   12  for 12-bit sample values
 * Only 8 and 12 are legal data precisions for lossy JPEG according to the
 * JPEG standard, and the IJG code does not support anything else!
 * We do not support run-time selection of data precision, sorry.
 */

immutable int BITS_IN_short  = 8;	/* use 8 or 12 */


/*
 * Maximum number of components (color channels) allowed in JPEG image.
 * To meet the letter of the JPEG spec, set this to 255.  However, darn
 * few applications need more than 4 channels (maybe 5 for CMYK + alpha
 * mask).  We recommend 10 as a reasonable compromise; use 4 if you are
 * really short on memory.  (Each allowed component costs a hundred or so
 * bytes of storage, whether actually used in an image or not.)
 */

immutable int MAX_COMPONENTS  = 10;	/* maximum number of image components */


/*
 * Basic data types.
 * You may need to change these if you have a machine with unusual data
 * type sizes; for example, "char" not 8 bits, "short" not 16 bits,
 * or "long" not 32 bits.  We don't care whether "int" is 16 or 32 bits,
 * but it had better be at least 16.
 */

/* Representation of a single sample (pixel element value).
 * We frequently allocate large arrays of these, so it's important to keep
 * them small.  But if you have memory to burn and access to char or short
 * arrays is very slow on your hardware, you might want to change these.
 */

/* short should be the smallest type that will hold the values 0..255.
 * You can use a signed char by having GETshort mask it with 0xFF.
 */
 
 

/* not HAVE_UNSIGNED_CHAR */

/* CHAR_IS_UNSIGNED */

/* HAVE_UNSIGNED_CHAR */

static if(BITS_IN_short == 8){
  immutable int MAXshort	= 255;
  immutable int CENTERshort	= 128;
}

/* BITS_IN_short == 8 */

/* short should be the smallest type that will hold the values 0..4095.
 * On nearly all machines "short" will do nicely.
 */

 static if(BITS_IN_short == 12){
  immutable int MAXshort	= 4095;
  immutable int CENTERshort	= 2048;
}
/* BITS_IN_short == 12 */


/* Representation of a DCT frequency coefficient.
 * This should be a signed value of at least 16 bits; "short" is usually OK.
 * Again, we allocate large arrays of these, but you can change to int
 * if you have memory to burn and "short" is really slow.
 */




/* Compressed datastreams are represented as arrays of char.
 * These must be EXACTLY 8 bits wide, at least once they are written to
 * external storage.  Note that when using the stdio data source/destination
 * managers, this is also the data type passed to fread/fwrite.
 */



/* not HAVE_UNSIGNED_CHAR */


/* CHAR_IS_UNSIGNED */

/* HAVE_UNSIGNED_CHAR */


/* These typedefs are used for various table entries and so forth.
 * They must be at least as wide as specified; but making them too big
 * won't cost a huge amount of memory, so we don't provide special
 * extraction code like we did for short.  (In other words, these
 * typedefs live at a different point on the speed/space tradeoff curve.)
 */

/* short must hold at least the values 0..255. */


/* not HAVE_UNSIGNED_CHAR */

/* not CHAR_IS_UNSIGNED */

/* CHAR_IS_UNSIGNED */
/* HAVE_UNSIGNED_CHAR */

/* Ushort must hold at least the values 0..65535. */


/* not HAVE_UNSIGNED_SHORT */

/* HAVE_UNSIGNED_SHORT */

/* short must hold at least the values -32768..32767. */

/* X11/xmd.h correctly defines short */


/* long must hold at least signed 32-bit values. */

/* X11/xmd.h correctly defines long */
/* Microsoft defines it in basetsd.h */
/* MinGW is slightly different */
/* Qt defines it in qglobal.h */


/* Datatype used for image dimensions.  The JPEG standard only supports
 * images up to 64K*64K due to 16-bit fields in SOF markers.  Therefore
 * "uint" is sufficient on all machines.  However, if you need to
 * handle larger images and you don't mind deviating from the spec, you
 * can change this datatype.
 */

immutable int JPEG_MAX_DIMENSION  = 65500;  /* a tad under 64K to prevent overflows */


/* These macros are used in all function definitions and extern declarations.
 * You could modify them if you need to change function linkage conventions;
 * in particular, you'll need to do that to make the library a Windows DLL.
 * Another application is to make all functions global for use with debuggers
 * or code profilers that require it.
 */

/* a function called through method pointers: */
/* a function used only in its module: */
/* a function referenced thru EXTERNs: */
/* a reference to a GLOBAL function: */


/* This macro is used to declare a "method", that is, a function pointer.
 * We want to supply prototype parameters if the compiler can cope.
 * Note that the arglist parameter must be parenthesized!
 * Again, you can customize this if you need special linkage keywords.
 */

/* The noreturn type identifier is used to declare functions
 * which cannot return.
 * Compilers can thus create more optimized code and perform
 * better checks for warnings and errors.
 * Static analyzer tools can make improved inferences about
 * execution paths and are prevented from giving false alerts.*
 
 * Unfortunately, the proposed specifications of corresponding
 * extensions in the Dec 2011 ISO C standard revision (C11),
 * GCC, MSVC, etc. are not viable.
 * Thus we introduce a user defined type to declare noreturn
 * functions at least for clarity.  A proper compiler would
 * have a suitable noreturn type to match in place of void.
 */




/* Here is the pseudo-keyword for declaring pointers that must be "far"
 * on 80x86 machines.  Most of the specialized coding for 80x86 is handled
 * by just saying "FAR* " where such a pointer is needed.  In a few places
 * explicit coding is needed; see uses of the NEED_FAR_POINTERS symbol.
 */

/*
 * On a few systems, type boolean and/or its values FALSE, TRUE may appear
 * in standard header files.  Or you may have conflicts with application-
 * specific header files that you want to include together with these files.
 * Defining HAVE_BOOLEAN before including jpeglib.h should make it work.
 */

			/* in case these macros already exist */
immutable int FALSE	= 0;		/* values of boolean */
immutable int TRUE	= 1;


/*
 * The remaining options affect code selection within the JPEG library,
 * but they don't need to be visible to most applications using the library.
 * To minimize application namespace pollution, the symbols won't be
 * defined unless JPEG_INTERNALS or JPEG_INTERNAL_OPTIONS has been defined.
 */



/*
 * These defines indicate whether to include various optional functions.
 * Undefining some of these symbols will produce a smaller but less capable
 * library.  Note that you can leave certain source files out of the
 * compilation/linking process if you've #undef'd the corresponding symbols.
 * (You may HAVE to do that if your compiler doesn't like null source files.)
 */

/* Capability options common to encoder and decoder: */

/* slow but accurate integer algorithm */
/* faster, less accurate integer method */
/* floating-point: accurate, fast on fast HW */

/* Encoder capability options: */
/* Multiple-scan JPEG files? */
/* Progressive JPEG? (Requires MULTISCAN)*/
/* Input rescaling via DCT? (Requires DCT_ISLOW)*/
/* Optimization of entropy coding parms? */
/* Note: if you selected 12-bit data precision, it is dangerous to turn off
 * ENTROPY_OPT_SUPPORTED.  The standard Huffman tables are only good for 8-bit
 * precision, so jchuff.c normally uses entropy optimization to compute
 * usable tables for higher precision.  If you don't want to do optimization,
 * you'll have to supply different default Huffman tables.
 * The exact same statements apply for progressive JPEG: the default tables
 * don't work for progressive mode.  (This may get fixed, however.)
 */
/* Input image smoothing option? */

/* Decoder capability options: */
/* Multiple-scan JPEG files? */
/* Progressive JPEG? (Requires MULTISCAN)*/
/* Output rescaling via IDCT? */
/* jpeg_save_markers() needed? */
/* Block smoothing? (Progressive only) */
/* Output rescaling at upsample stage? */
/* Fast path for sloppy upsampling? */
/* 1-pass color quantization? */
/* 2-pass color quantization? */

/* more capability options later, no doubt */


/*
 * Ordering of RGB data in scanlines passed to or from the application.
 * If your application wants to deal with data in the order B,G,R, just
 * change these macros.  You can also deal with formats such as R,G,B,X
 * (one extra byte per pixel) by changing RGB_PIXELSIZE.  Note that changing
 * the offsets will also change the order in which colormap data is organized.
 * RESTRICTIONS:
 * 1. The sample applications cjpeg,djpeg do NOT support modified RGB formats.
 * 2. The color quantizer modules will not behave desirably if RGB_PIXELSIZE
 *    is not 3 (they don't understand about dummy color components!).  So you
 *    can't use color quantization if you change that value.
 */

immutable int RGB_RED		= 0;	/* Offset of Red in an RGB scanline element */
immutable int RGB_GREEN	= 1;	/* Offset of Green */
immutable int RGB_BLUE	= 2;	/* Offset of Blue */
immutable int RGB_PIXELSIZE	= 3;	/* shorts per RGB scanline element */


/* Definitions for speed-related optimizations. */


/* If your compiler supports inline functions, define INLINE
 * as the inline keyword; otherwise define it as empty.
 */

/* default is to define it as empty */

/* On some machines (notably 68000 series) "int" is 32 bits, but multiplying
 * two 16-bit shorts is faster than multiplying two ints.  Define MULTIPLIER
 * as short on such a machine.  MULTIPLIER must be at least 16 bits wide.
 */

/* type for fastest integer multiply */


/* FAST_FLOAT should be either float or double, whichever is done faster
 * by your compiler.  (Note that this type is only used in the floating point
 * DCT routines, so it only matters if you've defined DCT_FLOAT_SUPPORTED.)
 * Typically, float is faster in ANSI C compilers, while double is faster in
 * pre-ANSI compilers (because they insist on converting to double anyway).
 * The code below therefore chooses float if we have ANSI-style prototypes.
 */

/* JPEG_INTERNAL_OPTIONS */
