module movietest;
//
import formats;
//

import libav.libavutil.avstring;
import libav.libavutil.avutil;
import libav.libavutil.dict;
import libav.libavutil.frame;
import libav.libavutil.imgutils;
import libav.libavutil.mathematics;
import libav.libavutil.mem;
import libav.libavutil.pixfmt;
import libav.libavutil.rational;

import libav.libavcodec.avcodec;

import libav.libavformat.avformat;
import libav.libavformat.avio;

import libav.libswscale.swscale;

//

import ngbase.loadimage;
import ngbase.bitmap;

//

import std.stdio;
import std.stream;
import core.stdc.stdlib;

alias std.stream.File File;

/**
 * フレームを画像として保存する
 * @param[in] frame      保存するフレーム
 * @param[in] filename  保存先のPATH
 */
void save_frame(string filename){
  AVCodec* codec;
  AVCodecContext* context;
  AVFormatContext* format;
  AVOutputFormat* output;
  AVStream* stream;
  AVDictionary* dict;
  uint i;
  int ret;
  int x;
  int y;
  int got_output;
  AVFrame* picture;
  AVPacket packet;
  //
  long pts;
  //
  codec = avcodec_find_encoder(AVCodecID.AV_CODEC_ID_HUFFYUV);
  if(!codec){
    stderr.writeln("codec not found");
    exit(1);
  }
  context = avcodec_alloc_context3(codec);
  picture = avcodec_alloc_frame();
  //
  format = avformat_alloc_context();
  output = av_guess_format(ptr(AVFormat.AV_FORMAT_AVI), null, null);
  stream = avformat_new_stream(format, codec);
  
  format.oformat = output;
  stream.codec = context;
  
  av_strlcpy(format.filename.ptr, filename.ptr, format.filename.length);
  /* put sample parameters */
  context.bit_rate = 400000;
  /* resolution must be a multiple of two */
  context.width = 256;
  context.height = 256;
  
  context.codec_id = codec.id;
  context.codec_type = AVMediaType.AVMEDIA_TYPE_VIDEO;
  /* frames per second */
  context.time_base = AVRational(1, 30);
  context.gop_size = 10; /* emit one intra frame every ten frames */
  context.max_b_frames = 1;
  //
  context.pix_fmt = AVPixelFormat.AV_PIX_FMT_BGRA;
  /* ピクセルフォーマットは、codecで対応されているかチェック */
  writefln("%d", context.pix_fmt);
  for(i = 0;; i += 1){
    writefln("%d", codec.pix_fmts[i]);
    if(context.pix_fmt == codec.pix_fmts[i]){
      break;
    }
    if(codec.pix_fmts[i] < 0){
      context.pix_fmt = codec.pix_fmts[0];
      break;
    }
  }
  //
  stream.time_base = context.time_base;
  /* open it */
  if(avcodec_open2(context, codec, null) < 0){
    stderr.writeln("could not open codec");
    exit(1);
  }
  /* 出力ファイルを開く */
  if (!(format.flags & AVFMT_NOFILE)){
    ret = avio_open(&format.pb, format.filename.ptr, AVIO_FLAG_WRITE);
    if(ret < 0){
      stderr.writeln("can't open output file.");
    }
  }
  
  ret = av_image_alloc(picture.data.ptr, picture.linesize.ptr, context.width, context.height, context.pix_fmt, 32);
  if(ret < 0){
    stderr.writeln("could not alloc raw picture buffer");
    exit(1);
  }
  picture.format = context.pix_fmt;
  picture.width  = context.width;
  picture.height = context.height;
  //
  avformat_write_header(format, &dict);
  writefln("%d %d %d %d", picture.linesize[0], picture.linesize[1], picture.linesize[2], picture.linesize[3]);
  /* encode 1 second of video */
  for(i = 0; i < 300; i++) {
    av_init_packet(&packet);
    packet.data = null;    // packet data will be allocated by the encoder
    packet.size = 0;
    
    /* prepare a dummy image */
    /* Y */
    for(y = 0; y < context.height; y++) {
      for(x = 0; x < context.width; x++) {
        picture.data[0][(y * context.width + x) * 4 + 0] = cast(ubyte)((x + y + i * 1) & 0xFF);
        picture.data[0][(y * context.width + x) * 4 + 1] = cast(ubyte)((x + y + i * 2) & 0xFF);
        picture.data[0][(y * context.width + x) * 4 + 2] = cast(ubyte)((x + y + i * 3) & 0xFF);
        picture.data[0][(y * context.width + x) * 4 + 3] = cast(ubyte)((x + y + i * 4) & 0xFF);
      }
      
    }
    
    picture.pts = i;
    /* encode the image */
    ret = avcodec_encode_video2(context, &packet, picture, &got_output);
    
    if(ret < 0){
      stderr.writeln("error encoding frame");
      exit(1);
    }
    if(got_output){
      if (packet.pts != AV_NOPTS_VALUE){
        packet.pts = av_rescale_q(packet.pts, context.time_base, stream.time_base);
      }
      if (packet.dts != AV_NOPTS_VALUE){
        packet.dts = av_rescale_q(packet.dts, context.time_base, stream.time_base);
      }
      writefln("encoding delay frame %3d (size=%5d)", i, packet.size);
      av_interleaved_write_frame(format, &packet);
      av_free_packet(&packet);
    }
  }
  
  /* get the delayed frames */
  for (got_output = 1; got_output; i++) {
    
    ret = avcodec_encode_video2(context, &packet, null, &got_output);
    if (ret < 0) {
      stderr.writeln("error encoding frame");
      exit(1);
    }
    
    if(got_output){
      packet.stream_index = 0;
      writefln("encoding frame %3d (size=%5d)", i, packet.size);
      av_interleaved_write_frame(format, &packet);
      av_free_packet(&packet);
    }
  }
  /* add sequence end code to have a real mpeg file */
  av_write_trailer(format);
  if((format.flags & AVFMT_NOFILE) != 0){
    ret = avio_close(format.pb);
    if(ret < 0){
      stderr.writeln("can't close output file.");
    }
  }
  
  avcodec_close(context);
  av_free(context);
  av_freep(&picture.data[0]);
  av_freep(&format.streams[0]);
  avcodec_free_frame(&picture);
  writeln("Finish!");
  return;
}

int main(string[] agev){
  ubyte* buffer;
  AVFrame *frame;
  //
  av_register_all();
  //
  if(agev.length > 1){
    save_frame(agev[1]);
  }else{
    save_frame("out.avi");
  }
  //
  
  //
  return 0;
}
